@@ -3,7 +3,7 @@ Implements the RTS generic Target fabric classes.
 
 This file is part of RTSLib.
 Copyright (c) 2011-2013 by Datera, Inc.
-Copyright (c) 2011-2013 by Red Hat, Inc.
+Copyright (c) 2011-2014 by Red Hat, Inc.
 
 Licensed under the Apache License, Version 2.0 (the "License"); you may
 not use this file except in compliance with the License. You may obtain
@@ -18,19 +18,22 @@ License for the specific language governing permissions and limitations
 under the License.
 '''
 
-import re
 import os
 from glob import iglob as glob
+from functools import partial
+from six.moves import range
 import uuid
 
-from node import CFSNode
-from os.path import isdir
-from utils import RTSLibError, RTSLibBrokenLink
-from utils import fread, fwrite, normalize_wwn, generate_wwn
-from utils import dict_remove, set_attributes, set_parameters, ignored
-from utils import _get_auth_attr, _set_auth_attr
-import tcm
-from functools import partial
+from .node import CFSNode
+from .utils import RTSLibBrokenLink, RTSLibError
+from .utils import fread, fwrite, normalize_wwn, generate_wwn
+from .utils import dict_remove, set_attributes, set_parameters, ignored
+from .utils import _get_auth_attr, _set_auth_attr
+from . import tcm
+
+import six
+
+auth_params = ('userid', 'password', 'mutual_userid', 'mutual_password')
 
 class Target(CFSNode):
     '''
@@ -156,7 +159,7 @@ class TPG(CFSNode):
         @param parent_target: The parent Target object of the TPG.
         @type parent_target: Target
         @param tag: The TPG Tag (TPGT).
-        @type tag: int > 0
+        @type tag: positive int
         @param mode:An optionnal string containing the object creation mode:
             - I{'any'} means the configFS object will be either looked up or
               created.
@@ -170,22 +173,22 @@ class TPG(CFSNode):
 
         if tag is None:
             tags = [tpg.tag for tpg in parent_target.tpgs]
-            for index in xrange(1048576):
+            for index in range(1048576):
                 if index not in tags and index > 0:
                     tag = index
                     break
             if tag is None:
-                raise RTSLibError("Cannot find an available TPG Tag.")
+                raise RTSLibError("Cannot find an available TPG Tag")
         else:
             tag = int(tag)
-            if not tag > 0:
-                raise RTSLibError("The TPG Tag must be >0.")
+            if tag < 0:
+                raise RTSLibError("The TPG Tag must be >=0")
         self._tag = tag
 
         if isinstance(parent_target, Target):
             self._parent_target = parent_target
         else:
-            raise RTSLibError("Invalid parent Target.")
+            raise RTSLibError("Invalid parent Target")
 
         self._path = "%s/tpgt_%d" % (self.parent_target.path, self.tag)
 
@@ -195,7 +198,7 @@ class TPG(CFSNode):
                 if filename.startswith("tpgt_") \
                    and os.path.isdir("%s/%s" % (target_path, filename)) \
                    and filename != "tpgt_%d" % self.tag:
-                    raise RTSLibError("Target cannot have multiple TPGs.")
+                    raise RTSLibError("Target cannot have multiple TPGs")
 
         self._create_in_cfs_ine(mode)
         if self.has_feature('nexus') and not self._get_nexus():
@@ -263,9 +266,9 @@ class TPG(CFSNode):
         self._check_self()
 
         if not self.has_feature('nexus'):
-            raise RTSLibError("The TPG does not use a nexus.")
+            raise RTSLibError("The TPG does not use a nexus")
         if self._get_nexus():
-            raise RTSLibError("The TPG's nexus initiator WWN is already set.")
+            raise RTSLibError("The TPG's nexus initiator WWN is already set")
 
         fm = self.parent_target.fabric_module
 
@@ -288,6 +291,20 @@ class TPG(CFSNode):
             fm = self.parent_target.fabric_module
             yield NodeACL(self, fm.from_fabric_wwn(node_acl_dir), 'lookup')
 
+    def _list_node_acl_groups(self):
+        self._check_self()
+        if not self.has_feature('acls'):
+            return
+
+        names = set([])
+
+        for na in self.node_acls:
+            tag = na.tag
+            if tag:
+                names.add(tag)
+
+        return (NodeACLGroup(self, n) for n in names)
+
     def _list_luns(self):
         self._check_self()
         lun_dirs = [os.path.basename(path)
@@ -366,6 +383,9 @@ class TPG(CFSNode):
     node_acls = property(_list_node_acls,
                          doc="Get the list of NodeACL objects currently " \
                          + "attached to the TPG.")
+    node_acl_groups = property(_list_node_acl_groups,
+                         doc="Get the list of NodeACL groups currently " \
+                         + "attached to the TPG.")
     luns = property(_list_luns,
                     doc="Get the list of LUN objects currently attached " \
                     + "to the TPG.")
@@ -400,8 +420,8 @@ class TPG(CFSNode):
     @classmethod
     def setup(cls, t_obj, tpg, err_func):
         tpg_obj = cls(t_obj, tag=tpg.get("tag", None))
-        set_attributes(tpg_obj, tpg.get('attributes', {}))
-        set_parameters(tpg_obj, tpg.get('parameters', {}))
+        set_attributes(tpg_obj, tpg.get('attributes', {}), err_func)
+        set_parameters(tpg_obj, tpg.get('parameters', {}), err_func)
 
         for lun in tpg.get('luns', []):
             LUN.setup(tpg_obj, lun, err_func)
@@ -415,7 +435,7 @@ class TPG(CFSNode):
         tpg_obj.enable = tpg.get('enable', True)
         dict_remove(tpg, ('luns', 'portals', 'node_acls', 'tag',
                           'attributes', 'parameters', 'enable'))
-        for name, value in tpg.iteritems():
+        for name, value in six.iteritems(tpg):
             if value:
                 try:
                     setattr(tpg_obj, name, value)
@@ -431,7 +451,7 @@ class TPG(CFSNode):
         d['portals'] = [portal.dump() for portal in self.network_portals]
         d['node_acls'] =  [acl.dump() for acl in self.node_acls]
         if self.has_feature("auth"):
-            for attr in ("userid", "password", "mutual_userid", "mutual_password"):
+            for attr in auth_params:
                 val = getattr(self, "chap_" + attr, None)
                 if val:
                     d["chap_" + attr] = val
@@ -444,7 +464,7 @@ class LUN(CFSNode):
     A LUN is identified by its parent TPG and LUN index.
     '''
 
-    MAX_LUN = 16383
+    MAX_TARGET_LUN = 65535
 
     # LUN private stuff
 
@@ -467,7 +487,7 @@ class LUN(CFSNode):
         @param parent_tpg: The parent TPG object.
         @type parent_tpg: TPG
         @param lun: The LUN index.
-        @type lun: 0-255
+        @type lun: 0-65535
         @param storage_object: The storage object to be exported as a LUN.
         @type storage_object: StorageObject subclass
         @param alias: An optional parameter to manually specify the LUN alias.
@@ -480,20 +500,20 @@ class LUN(CFSNode):
         if isinstance(parent_tpg, TPG):
             self._parent_tpg = parent_tpg
         else:
-            raise RTSLibError("Invalid parent TPG.")
+            raise RTSLibError("Invalid parent TPG")
 
         if lun is None:
             luns = [l.lun for l in self.parent_tpg.luns]
-            for index in xrange(self.MAX_LUN+1):
+            for index in range(self.MAX_TARGET_LUN+1):
                 if index not in luns:
                     lun = index
                     break
             if lun is None:
-                raise RTSLibError("All LUNs 0-%d in use" % self.MAX_LUN)
+                raise RTSLibError("All LUNs 0-%d in use" % self.MAX_TARGET_LUN)
         else:
             lun = int(lun)
-            if lun < 0 or lun > self.MAX_LUN:
-                raise RTSLibError("LUN must be 0 to %d" % self.MAX_LUN)
+            if lun < 0 or lun > self.MAX_TARGET_LUN:
+                raise RTSLibError("LUN must be 0 to %d" % self.MAX_TARGET_LUN)
 
         self._lun = lun
 
@@ -501,7 +521,7 @@ class LUN(CFSNode):
 
         if storage_object is None and alias is not None:
             raise RTSLibError("The alias parameter has no meaning " \
-                              + "without the storage_object parameter.")
+                              + "without the storage_object parameter")
 
         if storage_object is not None:
             self._create_in_cfs_ine('create')
@@ -527,33 +547,21 @@ class LUN(CFSNode):
         if storage_object.exists:
             source = storage_object.path
         else:
-            raise RTSLibError("storage_object does not exist in configFS.")
+            raise RTSLibError("storage_object does not exist in configFS")
 
         os.symlink(source, destination)
 
     def _get_alias(self):
         self._check_self()
-        alias = None
         for path in os.listdir(self.path):
             if os.path.islink("%s/%s" % (self.path, path)):
-                alias = os.path.basename(path)
-                break
-        if alias is None:
-            raise RTSLibBrokenLink("Broken LUN in configFS, no " \
-                                         + "storage object attached.")
-        else:
-            return alias
+                return os.path.basename(path)
+
+        raise RTSLibBrokenLink("Broken LUN in configFS, no storage object")
 
     def _get_storage_object(self):
         self._check_self()
-        alias_path = None
-        for path in os.listdir(self.path):
-            if os.path.islink("%s/%s" % (self.path, path)):
-                alias_path = os.path.realpath("%s/%s" % (self.path, path))
-                break
-        if alias_path is None:
-            raise RTSLibBrokenLink("Broken LUN in configFS, no "
-                                   + "storage object attached.")
+        alias_path = os.path.realpath("%s/%s" % (self.path, self.alias))
         return tcm.StorageObject.so_from_path(alias_path)
 
     def _get_parent_tpg(self):
@@ -571,9 +579,47 @@ class LUN(CFSNode):
 
         for na in tpg.node_acls:
             for mlun in na.mapped_luns:
-                if os.path.realpath("%s/%s" % (mlun.path, mlun._get_alias())) == self.path:
+                if os.path.realpath("%s/%s" % (mlun.path, mlun.alias)) == self.path:
                     yield mlun
 
+
+    # pass through backends will not have setup all the default
+    # ALUA structs in the kernel. If the kernel has been setup,
+    # a user created group or default_tg_pt_gp will be returned.
+    # If the kernel was not properly setup an empty string is
+    # return in alua_tg_pt_gp. Writing to alua_tg_pt_gp will crash
+    # older kernels and will return a -Exyz code in newer ones.
+    def _get_alua_tg_pt_gp_name(self):
+        self._check_self()
+
+        storage_object = self._get_storage_object()
+        if storage_object.alua_supported is False:
+            return None
+
+        path = "%s/alua_tg_pt_gp" % self.path
+        try:
+            info = fread(path)
+            if not info:
+                return None
+            group_line = info.splitlines()[0]
+            return group_line.split(':')[1].strip()
+        except IOError as e:
+            return None
+
+    def _set_alua_tg_pt_gp_name(self, group_name):
+        self._check_self()
+
+        if not self._get_alua_tg_pt_gp_name():
+            return -1
+
+        path = "%s/alua_tg_pt_gp" % self.path
+        try:
+            fwrite(path, group_name)
+        except IOError as e:
+            return -1
+
+        return 0
+
     # LUN public stuff
 
     def delete(self):
@@ -607,6 +653,8 @@ class LUN(CFSNode):
             doc="Get the LUN alias.")
     mapped_luns = property(_list_mapped_luns,
             doc="List all MappedLUN objects referencing this LUN.")
+    alua_tg_pt_gp_name = property(_get_alua_tg_pt_gp_name, _set_alua_tg_pt_gp_name,
+            doc="Get and Set the LUN's ALUA Target Port Group")
 
     @classmethod
     def setup(cls, tpg_obj, lun, err_func):
@@ -629,16 +677,24 @@ class LUN(CFSNode):
             return
 
         try:
-            cls(tpg_obj, lun['index'], storage_object=match_so)
+           lun_obj =  cls(tpg_obj, lun['index'], storage_object=match_so, alias=lun.get('alias'))
         except (RTSLibError, KeyError):
             err_func("Creating TPG %d LUN index %d failed" %
                      (tpg_obj.tag, lun['index']))
 
+        try:
+            lun_obj.alua_tg_pt_gp_name = lun['alua_tg_pt_gp_name']
+        except KeyError:
+            # alua_tg_pt_gp support not present in older versions
+            pass
+
     def dump(self):
         d = super(LUN, self).dump()
         d['storage_object'] = "/backstores/%s/%s" % \
             (self.storage_object.plugin,  self.storage_object.name)
         d['index'] = self.lun
+        d['alias'] = self.alias
+        d['alua_tg_pt_gp_name'] = self.alua_tg_pt_gp_name
         return d
 
 
@@ -680,12 +736,12 @@ class NetworkPortal(CFSNode):
         try:
             self._port = int(port)
         except ValueError:
-            raise RTSLibError("Invalid port.")
+            raise RTSLibError("Invalid port")
 
         if isinstance(parent_tpg, TPG):
             self._parent_tpg = parent_tpg
         else:
-            raise RTSLibError("Invalid parent TPG.")
+            raise RTSLibError("Invalid parent TPG")
 
         self._path = "%s/np/%s:%d" \
             % (self.parent_tpg.path, self.ip_address, self.port)
@@ -693,7 +749,7 @@ class NetworkPortal(CFSNode):
         try:
             self._create_in_cfs_ine(mode)
         except OSError as msg:
-            raise RTSLibError(msg[1])
+            raise RTSLibError(msg)
 
     def _get_ip_address(self):
         return self._ip_address
@@ -719,10 +775,27 @@ class NetworkPortal(CFSNode):
             if os.path.isfile(path):
                 raise RTSLibError("Cannot change iser")
 
+    def _get_offload(self):
+        try:
+            # only offload at the moment is cxgbit
+            return bool(int(fread("%s/cxgbit" % self.path)))
+        except IOError:
+            return False
+
+    def _set_offload(self, boolean):
+        path = "%s/cxgbit" % self.path
+        try:
+            fwrite(path, str(int(boolean)))
+        except IOError:
+            # b/w compat: don't complain if cxgbit entry is missing
+            if os.path.isfile(path):
+                raise RTSLibError("Cannot change offload")
+
     # NetworkPortal public stuff
 
     def delete(self):
         self.iser = False
+        self.offload = False
         super(NetworkPortal, self).delete()
 
     parent_tpg = property(_get_parent_tpg,
@@ -734,6 +807,9 @@ class NetworkPortal(CFSNode):
     iser = property(_get_iser, _set_iser,
                     doc="Get or set a boolean value representing if this " \
                         + "NetworkPortal supports iSER.")
+    offload = property(_get_offload, _set_offload,
+                    doc="Get or set a boolean value representing if this " \
+                        + "NetworkPortal supports offload.")
 
     @classmethod
     def setup(cls, tpg_obj, p, err_func):
@@ -747,15 +823,17 @@ class NetworkPortal(CFSNode):
         try:
             np = cls(tpg_obj, p['ip_address'], p['port'])
             np.iser = p.get('iser', False)
-        except (RTSLibError, KeyError):
-            err_func("Creating NetworkPortal object %s:%s failed" %
-                     (p['ip_address'], p['port']))
+            np.offload = p.get('offload', False)
+        except (RTSLibError, KeyError) as e:
+            err_func("Creating NetworkPortal object %s:%s failed: %s" %
+                     (p['ip_address'], p['port'], e))
 
     def dump(self):
         d = super(NetworkPortal, self).dump()
         d['port'] = self.port
         d['ip_address'] = self.ip_address
         d['iser'] = self.iser
+        d['offload'] = self.offload
         return d
 
 
@@ -791,7 +869,7 @@ class NodeACL(CFSNode):
         if isinstance(parent_tpg, TPG):
             self._parent_tpg = parent_tpg
         else:
-            raise RTSLibError("Invalid parent TPG.")
+            raise RTSLibError("Invalid parent TPG")
 
         fm = self.parent_tpg.parent_target.fabric_module
         self._node_wwn, self.wwn_type = normalize_wwn(fm.wwn_types, node_wwn)
@@ -945,13 +1023,13 @@ class NodeACL(CFSNode):
             err_func("Error when creating NodeACL for %s: %s" % (acl['node_wwn'], e))
             return
 
-        set_attributes(acl_obj, acl.get('attributes', {}))
+        set_attributes(acl_obj, acl.get('attributes', {}), err_func)
 
         for mlun in acl.get('mapped_luns', []):
             MappedLUN.setup(tpg_obj, acl_obj, mlun, err_func)
 
         dict_remove(acl, ('attributes', 'mapped_luns', 'node_wwn'))
-        for name, value in acl.iteritems():
+        for name, value in six.iteritems(acl):
             if value:
                 try:
                     setattr(acl_obj, name, value)
@@ -966,7 +1044,7 @@ class NodeACL(CFSNode):
         if self.tag:
             d['tag'] = self.tag
         if self.has_feature("auth"):
-            for attr in ("userid", "password", "mutual_userid", "mutual_password"):
+            for attr in auth_params:
                 val = getattr(self, "chap_" + attr, None)
                 if val:
                     d["chap_" + attr] = val
@@ -982,13 +1060,17 @@ class MappedLUN(CFSNode):
     the initiator node as the MappedLUN.
     '''
 
+    MAX_LUN = 255
+
     # MappedLUN private stuff
 
     def __repr__(self):
-        return "<MappedLUN %d -> %d>" % (self.mapped_lun, self.tpg_lun.lun)
+        return "<MappedLUN %s lun %d -> tpg%d lun %d>" % \
+            (self.parent_nodeacl.node_wwn, self.mapped_lun,
+             self.parent_nodeacl.parent_tpg.tag, self.tpg_lun.lun)
 
     def __init__(self, parent_nodeacl, mapped_lun,
-                 tpg_lun=None, write_protect=None):
+                 tpg_lun=None, write_protect=None, alias=None):
         '''
         A MappedLUN object can be instanciated in two ways:
             - B{Creation mode}: If I{tpg_lun} is specified, the underlying
@@ -1015,35 +1097,38 @@ class MappedLUN(CFSNode):
 
         if not isinstance(parent_nodeacl, NodeACL):
             raise RTSLibError("The parent_nodeacl parameter must be " \
-                              + "a NodeACL object.")
+                              + "a NodeACL object")
         else:
             self._parent_nodeacl = parent_nodeacl
             if not parent_nodeacl.exists:
-                raise RTSLibError("The parent_nodeacl does not exist.")
+                raise RTSLibError("The parent_nodeacl does not exist")
 
         try:
             self._mapped_lun = int(mapped_lun)
         except ValueError:
             raise RTSLibError("The mapped_lun parameter must be an " \
-                              + "integer value.")
+                              + "integer value")
+
+        if self._mapped_lun < 0 or self._mapped_lun > self.MAX_LUN:
+            raise RTSLibError("Mapped LUN must be 0 to %d" % self.MAX_LUN)
 
         self._path = "%s/lun_%d" % (self.parent_nodeacl.path, self.mapped_lun)
 
         if tpg_lun is None and write_protect is not None:
             raise RTSLibError("The write_protect parameter has no " \
-                              + "meaning without the tpg_lun parameter.")
+                              + "meaning without the tpg_lun parameter")
 
         if tpg_lun is not None:
             self._create_in_cfs_ine('create')
             try:
-                self._configure(tpg_lun, write_protect)
+                self._configure(tpg_lun, write_protect, alias)
             except:
                 self.delete()
                 raise
         else:
             self._create_in_cfs_ine('lookup')
 
-    def _configure(self, tpg_lun, write_protect):
+    def _configure(self, tpg_lun, write_protect, alias):
         self._check_self()
         if isinstance(tpg_lun, LUN):
             tpg_lun = tpg_lun.lun
@@ -1052,17 +1137,19 @@ class MappedLUN(CFSNode):
                 tpg_lun = int(tpg_lun)
             except ValueError:
                 raise RTSLibError("The tpg_lun must be either an "
-                                  + "integer or a LUN object.")
+                                  + "integer or a LUN object")
         # Check that the tpg_lun exists in the TPG
         for lun in self.parent_nodeacl.parent_tpg.luns:
             if lun.lun == tpg_lun:
                 tpg_lun = lun
                 break
         if not (isinstance(tpg_lun, LUN) and tpg_lun):
-            raise RTSLibError("LUN %s does not exist in this TPG."
+            raise RTSLibError("LUN %s does not exist in this TPG"
                               % str(tpg_lun))
-        os.symlink(tpg_lun.path, "%s/%s"
-                   % (self.path, str(uuid.uuid4())[-10:]))
+
+        if not alias:
+            alias = str(uuid.uuid4())[-10:]
+        os.symlink(tpg_lun.path, "%s/%s" % (self.path, alias))
 
         try:
             self.write_protect = int(write_protect) > 0
@@ -1071,16 +1158,11 @@ class MappedLUN(CFSNode):
 
     def _get_alias(self):
         self._check_self()
-        alias = None
         for path in os.listdir(self.path):
             if os.path.islink("%s/%s" % (self.path, path)):
-                alias = os.path.basename(path)
-                break
-        if alias is None:
-            raise RTSLibBrokenLink("Broken LUN in configFS, no " \
-                                         + "storage object attached.")
-        else:
-            return alias
+                return os.path.basename(path)
+
+        raise RTSLibBrokenLink("Broken LUN in configFS, no storage object")
 
     def _get_mapped_lun(self):
         return self._mapped_lun
@@ -1103,12 +1185,12 @@ class MappedLUN(CFSNode):
 
     def _get_tpg_lun(self):
         self._check_self()
-        path = os.path.realpath("%s/%s" % (self.path, self._get_alias()))
+        path = os.path.realpath("%s/%s" % (self.path, self.alias))
         for lun in self.parent_nodeacl.parent_tpg.luns:
             if lun.path == path:
                 return lun
 
-        raise RTSLibBrokenLink("Broken MappedLUN, no TPG LUN found !")
+        raise RTSLibBrokenLink("Broken MappedLUN, no TPG LUN found")
 
     def _get_node_wwn(self):
         self._check_self()
@@ -1122,7 +1204,7 @@ class MappedLUN(CFSNode):
         '''
         self._check_self()
         try:
-            lun_link = "%s/%s" % (self.path, self._get_alias())
+            lun_link = "%s/%s" % (self.path, self.alias)
         except RTSLibBrokenLink:
             pass
         else:
@@ -1140,6 +1222,8 @@ class MappedLUN(CFSNode):
             doc="Get the TPG LUN object the MappedLUN is pointing at.")
     node_wwn = property(_get_node_wwn,
             doc="Get the wwn of the node for which the TPG LUN is mapped.")
+    alias = property(_get_alias,
+            doc="Get the MappedLUN alias.")
 
     @classmethod
     def setup(cls, tpg_obj, acl_obj, mlun, err_func):
@@ -1163,7 +1247,8 @@ class MappedLUN(CFSNode):
 
         try:
             mlun_obj = cls(acl_obj, mlun['index'],
-                           tpg_lun_obj, mlun.get('write_protect'))
+                           tpg_lun_obj, mlun.get('write_protect'),
+                           mlun.get('alias'))
             mlun_obj.tag = mlun.get("tag", None)
         except (RTSLibError, KeyError):
             err_func("Creating MappedLUN object %d failed" % mlun['index'])
@@ -1173,9 +1258,330 @@ class MappedLUN(CFSNode):
         d['write_protect'] = self.write_protect
         d['index'] = self.mapped_lun
         d['tpg_lun'] = self.tpg_lun.lun
+        d['alias'] = self.alias
         return d
 
 
+class Group(object):
+    '''
+    An abstract base class akin to CFSNode, but for classes that
+    emulate a higher-level group object across the actual NodeACL
+    configfs structure.
+    '''
+    def __init__(self, members_func):
+        '''
+        members_func is a function that takes a self argument
+        and returns an iterator of the objects that the
+        derived Group class is grouping.
+        '''
+        self._mem_func = members_func
+
+    def _get_first_member(self):
+        try:
+            return next(self._mem_func(self))
+        except StopIteration:
+            raise IndexError("Group is empty")
+
+    def _get_prop(self, prop):
+        '''
+        Helper fn to use with partial() to support getting a
+        property value from the first member of the group.
+        (All members of the group should be identical.)
+        '''
+        return getattr(self._get_first_member(), prop)
+
+    def _set_prop(self, value, prop):
+        '''
+        Helper fn to use with partial() to support setting a
+        property value in all members of the group.
+
+        Caution: Arguments reversed!
+        This is so partial() can be used on property name.
+        '''
+
+        for mem in self._mem_func(self):
+            setattr(mem, prop, value)
+
+    def list_attributes(self, writable=None, readable=None):
+        return self._get_first_member().list_attributes(writable, readable)
+
+    def list_parameters(self, writable=None, readable=None):
+        return self._get_first_member().list_parameters(writable, readable)
+
+    def set_attribute(self, attribute, value):
+        for obj in self._mem_func(self):
+            obj.set_attribute(attribute, value)
+
+    def set_parameter(self, parameter, value):
+        for obj in self._mem_func(self):
+            obj.set_parameter(parameter, value)
+
+    def get_attribute(self, attribute):
+        return self._get_first_member().get_attribute(attribute)
+
+    def get_parameter(self, parameter):
+        return self._get_first_member().get_parameter(parameter)
+
+    def delete(self):
+        '''
+        Delete all members of the group.
+        '''
+        for mem in self._mem_func(self):
+            mem.delete()
+
+    @property
+    def exists(self):
+        return any(self._mem_func(self))
+
+
+def _check_group_name(name):
+    # Since all WWNs have a '.' in them, let's avoid confusion.
+    if '.' in name:
+        raise RTSLibError("'.' not permitted in group names.")
+
+
+class NodeACLGroup(Group):
+    '''
+    Allow a group of NodeACLs that share a tag to be managed collectively.
+    '''
+    def __repr__(self):
+        return "<NodeACLGroup %s>" % self.name
+
+    def __init__(self, parent_tpg, name):
+        super(NodeACLGroup, self).__init__(NodeACLGroup._node_acls.fget)
+        _check_group_name(name)
+        self._name = name
+        self._parent_tpg = parent_tpg
+
+    def _get_name(self):
+        return self._name
+
+    def _set_name(self, name):
+        _check_group_name(name)
+        for na in self._node_acls:
+            na.tag = name
+        self._name = name
+
+    @property
+    def parent_tpg(self):
+        '''
+        Get the parent TPG object.
+        '''
+        return self._parent_tpg
+
+    def add_acl(self, node_wwn):
+        '''
+        Add a WWN to the NodeACLGroup. If a NodeACL already exists for this WWN,
+        its configuration will be changed to match the NodeACLGroup, except for its
+        auth parameters, which can vary among group members.
+        @param node_wwn: An initiator WWN
+        @type node_wwn: string
+        '''
+        nacl = NodeACL(self.parent_tpg, node_wwn)
+
+        if nacl in self._node_acls:
+            return
+
+        # if joining a group, take its config
+        try:
+            model = next(self._node_acls)
+        except StopIteration:
+            pass
+        else:
+            for mlun in nacl.mapped_luns:
+                mlun.delete()
+
+            for mlun in model.mapped_luns:
+                MappedLUN(nacl, mlun.mapped_lun, mlun.tpg_lun, mlun.write_protect)
+
+            for item in model.list_attributes(writable=True):
+                nacl.set_attribute(item, model.get_attribute(item))
+            for item in model.list_parameters(writable=True):
+                nacl.set_parameter(item, model.get_parameter(item))
+        finally:
+            nacl.tag = self.name
+
+    def remove_acl(self, node_wwn):
+        '''
+        Remove a WWN from the NodeACLGroup.
+        @param node_wwn: An initiator WWN
+        @type node_wwn: string
+        '''
+        nacl = NodeACL(self.parent_tpg, node_wwn, mode='lookup')
+
+        nacl.delete()
+
+    @property
+    def _node_acls(self):
+        '''
+        Gives access to the underlying NodeACLs within this group.
+        '''
+        for na in self.parent_tpg.node_acls:
+            if na.tag == self.name:
+                yield na
+
+    @property
+    def wwns(self):
+        '''
+        Give the Node WWNs of members of this group.
+        '''
+        return (na.node_wwn for na in self._node_acls)
+
+    def has_feature(self, feature):
+        '''
+        Whether or not this NodeACL has a certain feature.
+        '''
+        return self._parent_tpg.has_feature(feature)
+
+    @property
+    def sessions(self):
+        '''
+        Yields any current sessions.
+        '''
+        for na in self._node_acls:
+            session = na.session
+            if session:
+                yield session
+
+    def mapped_lun_group(self, mapped_lun, tpg_lun=None, write_protect=None):
+        '''
+        Add a mapped lun to all group members.
+        '''
+        return MappedLUNGroup(self, mapped_lun=mapped_lun, tpg_lun=tpg_lun,
+                      write_protect=write_protect)
+
+    @property
+    def mapped_lun_groups(self):
+        '''
+        Generates all MappedLUNGroup objects in this NodeACLGroup.
+        '''
+        try:
+            first = self._get_first_member()
+        except IndexError:
+            return
+
+        for mlun in first.mapped_luns:
+            yield MappedLUNGroup(self, mlun.mapped_lun)
+
+    name = property(_get_name, _set_name,
+                    doc="Get/set NodeACLGroup name.")
+
+    def _get_chap(self, name):
+        for na in self._node_acls:
+            yield (na.node_wwn, getattr(na, "chap_" + name))
+
+    def _set_chap(self, name, value, wwn):
+        for na in self._node_acls:
+            if not wwn:
+                setattr(na, "chap_" + name, value)
+            elif wwn == na.node_wwn:
+                setattr(na, "chap_" + name, value)
+                break
+
+    def get_userids(self):
+        '''
+        Returns a (wwn, userid) tuple for each member of the group.
+        '''
+        return self._get_chap(name="userid")
+
+    def set_userids(self, value, wwn=None):
+        '''
+        If wwn, set the userid for just that wwn, otherwise set it for
+        all group members.
+        '''
+        return self._set_chap("userid", value, wwn)
+
+    def get_passwords(self):
+        '''
+        Returns a (wwn, password) tuple for each member of the group.
+        '''
+        return self._get_chap(name="password")
+
+    def set_passwords(self, value, wwn=None):
+        '''
+        If wwn, set the password for just that wwn, otherwise set it for
+        all group members.
+        '''
+        return self._set_chap("password", value, wwn)
+
+    def get_mutual_userids(self):
+        '''
+        Returns a (wwn, mutual_userid) tuple for each member of the group.
+        '''
+        return self._get_chap(name="mutual_userid")
+
+    def set_mutual_userids(self, value, wwn=None):
+        '''
+        If wwn, set the mutual_userid for just that wwn, otherwise set it for
+        all group members.
+        '''
+        return self._set_chap("mutual_userid", value, wwn)
+
+    def get_mutual_passwords(self):
+        '''
+        Returns a (wwn, mutual_password) tuple for each member of the group.
+        '''
+        return self._get_chap(name="mutual_password")
+
+    def set_mutual_passwords(self, value, wwn=None):
+        '''
+        If wwn, set the mutual_password for just that wwn, otherwise set it for
+        all group members.
+        '''
+        return self._set_chap("mutual_password", value, wwn)
+
+    tcq_depth = property(partial(Group._get_prop, prop="tcq_depth"),
+                         partial(Group._set_prop, prop="tcq_depth"),
+                         doc="Set or get the TCQ depth for the initiator "
+                         + "sessions matching this NodeACLGroup")
+    authenticate_target = property(partial(Group._get_prop, prop="authenticate_target"),
+                                   doc="Get the boolean authenticate target flag.")
+
+
+class MappedLUNGroup(Group):
+    '''
+    Used with NodeACLGroup, this aggregates all MappedLUNs with the same LUN
+    so that it can be configured across all members of the NodeACLGroup.
+    '''
+
+    def __repr__(self):
+        return "<MappedLUNGroup %s:lun %d>" % (self._nag.name, self._mapped_lun)
+
+    def __init__(self, nodeaclgroup, mapped_lun, *args, **kwargs):
+        super(MappedLUNGroup, self).__init__(MappedLUNGroup._mapped_luns.fget)
+        self._nag = nodeaclgroup
+        self._mapped_lun = mapped_lun
+        for na in self._nag._node_acls:
+            MappedLUN(na, mapped_lun=mapped_lun, *args, **kwargs)
+
+    @property
+    def _mapped_luns(self):
+        for na in self._nag._node_acls:
+            for mlun in na.mapped_luns:
+                if mlun.mapped_lun == self.mapped_lun:
+                    yield mlun
+
+    @property
+    def mapped_lun(self):
+        '''
+        Get the integer MappedLUN mapped_lun index.
+        '''
+        return self._mapped_lun
+
+    @property
+    def parent_nodeaclgroup(self):
+        '''
+        Get the parent NodeACLGroup object.
+        '''
+        return self._nag
+
+    write_protect = property(partial(Group._get_prop, prop="write_protect"),
+                             partial(Group._set_prop, prop="write_protect"),
+                             doc="Get or set the boolean write protection.")
+    tpg_lun = property(partial(Group._get_prop, prop="tpg_lun"),
+                       doc="Get the TPG LUN object the MappedLUN is pointing at.")
+
+
 def _test():
     from doctest import testmod
     testmod()
