@@ -22,6 +22,7 @@ import os
 import stat
 import re
 import glob
+import fcntl
 import resource
 from six.moves import range
 
@@ -32,6 +33,10 @@ from .utils import convert_scsi_path_to_hctl, convert_scsi_hctl_to_path
 from .utils import is_dev_in_use, get_blockdev_type
 from .utils import get_size_for_blk_dev, get_size_for_disk_name
 
+lock_file = '/var/run/rtslib_backstore.lock'
+if not os.path.exists('/var/run'):
+    os.makedirs('/var/run')
+
 def storage_object_get_alua_support_attr(so):
     '''
     Helper function that can be called by passthrough type of backends.
@@ -56,14 +61,14 @@ class StorageObject(CFSNode):
     def __repr__(self):
         return "<%s %s/%s>" % (self.__class__.__name__, self.plugin, self.name)
 
-    def __init__(self, name, mode):
+    def __init__(self, name, mode, index=None):
         super(StorageObject, self).__init__()
         if "/" in name or " " in name or "\t" in name or "\n" in name:
             raise RTSLibError("A storage object's name cannot contain "
                               " /, newline or spaces/tabs")
         else:
             self._name = name
-        self._backstore = _Backstore(name, type(self), mode)
+        self._backstore = _Backstore(name, type(self), mode, index)
         self._path = "%s/%s" % (self._backstore.path, self.name)
         self.plugin = self._backstore.plugin
         try:
@@ -126,8 +131,8 @@ class StorageObject(CFSNode):
         Build a StorageObject of the correct type from a configfs path.
         '''
         so_name = os.path.basename(path)
-        so_type = path.split("/")[-2].rsplit("_", 1)[0]
-        return so_mapping[so_type](so_name)
+        so_type, so_index = path.split("/")[-2].rsplit("_", 1)
+        return so_mapping[so_type](so_name, index=so_index)
 
     def _get_wwn(self):
         self._check_self()
@@ -184,8 +189,11 @@ class StorageObject(CFSNode):
     def _parse_info(self, key):
         self._check_self()
         info = fread("%s/info" % self.path)
-        return re.search(".*%s: ([^: ]+).*" \
-                         % key, ' '.join(info.split())).group(1)
+        try:
+            return re.search(".*%s: ([^: ]+).*" \
+                             % key, ' '.join(info.split())).group(1)
+        except AttributeError:
+            return None
 
     def _get_status(self):
         self._check_self()
@@ -248,7 +256,7 @@ class StorageObject(CFSNode):
 
     # StorageObject public stuff
 
-    def delete(self):
+    def delete(self, save=False):
         '''
         Recursively deletes a StorageObject object.
         This will delete all attached LUNs currently using the StorageObject
@@ -271,21 +279,23 @@ class StorageObject(CFSNode):
 
         super(StorageObject, self).delete()
         self._backstore.delete()
+        if save:
+            from .root import RTSRoot, default_save_file
+            RTSRoot().save_to_file(default_save_file, '/backstores/' + self.plugin  + '/' + self._name)
 
     def is_configured(self):
         '''
         @return: True if the StorageObject is configured, else returns False
         '''
-
         self._check_self()
         path = "%s/enable" % self.path
-        try:
-            configured = fread(path)
-        except IOError:
+        # If the StorageObject does not have the enable attribute,
+        # then it is always enabled.
+        if os.path.isfile(path):
+            return bool(int(fread(path)))
+        else:
             return True
 
-        return bool(int(configured))
-
     version = property(_get_version,
             doc="Get the version of the StorageObject's backstore")
     name = property(_get_name,
@@ -319,7 +329,7 @@ class PSCSIStorageObject(StorageObject):
 
     # PSCSIStorageObject private stuff
 
-    def __init__(self, name, dev=None):
+    def __init__(self, name, dev=None, index=None):
         '''
         A PSCSIStorageObject can be instantiated in two ways:
             - B{Creation mode}: If I{dev} is specified, the underlying configFS
@@ -341,14 +351,14 @@ class PSCSIStorageObject(StorageObject):
         @return: A PSCSIStorageObject object.
         '''
         if dev is not None:
-            super(PSCSIStorageObject, self).__init__(name, 'create')
+            super(PSCSIStorageObject, self).__init__(name, 'create', index)
             try:
                 self._configure(dev)
             except:
                 self.delete()
                 raise
         else:
-            super(PSCSIStorageObject, self).__init__(name, 'lookup')
+            super(PSCSIStorageObject, self).__init__(name, 'lookup', index)
 
     def _configure(self, dev):
         self._check_self()
@@ -471,7 +481,7 @@ class RDMCPStorageObject(StorageObject):
 
     # RDMCPStorageObject private stuff
 
-    def __init__(self, name, size=None, wwn=None, nullio=False):
+    def __init__(self, name, size=None, wwn=None, nullio=False, index=None):
         '''
         A RDMCPStorageObject can be instantiated in two ways:
             - B{Creation mode}: If I{size} is specified, the underlying
@@ -496,14 +506,14 @@ class RDMCPStorageObject(StorageObject):
         '''
 
         if size is not None:
-            super(RDMCPStorageObject, self).__init__(name, 'create')
+            super(RDMCPStorageObject, self).__init__(name, 'create', index)
             try:
                 self._configure(size, wwn, nullio)
             except:
                 self.delete()
                 raise
         else:
-            super(RDMCPStorageObject, self).__init__(name, 'lookup')
+            super(RDMCPStorageObject, self).__init__(name, 'lookup', index)
 
     def _configure(self, size, wwn, nullio):
         self._check_self()
@@ -569,7 +579,7 @@ class FileIOStorageObject(StorageObject):
     # FileIOStorageObject private stuff
 
     def __init__(self, name, dev=None, size=None,
-                 wwn=None, write_back=False):
+                 wwn=None, write_back=False, aio=False, index=None):
         '''
         A FileIOStorageObject can be instantiated in two ways:
             - B{Creation mode}: If I{dev} and I{size} are specified, the
@@ -601,16 +611,16 @@ class FileIOStorageObject(StorageObject):
         '''
 
         if dev is not None:
-            super(FileIOStorageObject, self).__init__(name, 'create')
+            super(FileIOStorageObject, self).__init__(name, 'create', index)
             try:
-                self._configure(dev, size, wwn, write_back)
+                self._configure(dev, size, wwn, write_back, aio)
             except:
                 self.delete()
                 raise
         else:
-            super(FileIOStorageObject, self).__init__(name, 'lookup')
+            super(FileIOStorageObject, self).__init__(name, 'lookup', index)
 
-    def _configure(self, dev, size, wwn, write_back):
+    def _configure(self, dev, size, wwn, write_back, aio):
         self._check_self()
         block_type = get_blockdev_type(dev)
         if block_type is None: # a file
@@ -638,6 +648,9 @@ class FileIOStorageObject(StorageObject):
             self.set_attribute("emulate_write_cache", 1)
             self._control("fd_buffered_io=%d" % write_back)
 
+        if aio:
+            self._control("fd_async_io=%d" % aio)
+
         self._set_udev_path(dev)
 
         self._enable()
@@ -660,6 +673,15 @@ class FileIOStorageObject(StorageObject):
     def _is_block(self):
         return get_blockdev_type(self.udev_path) is not None
 
+    def _aio(self):
+        self._check_self()
+        info = fread("%s/info" % self.path)
+        r = re.search(".*Async: ([^: ]+).*", ' '.join(info.split()))
+        if not r:  # for backward compatibility with old kernels
+            return False
+
+        return bool(int(r.group(1)))
+
     # FileIOStorageObject public stuff
 
     write_back = property(_get_wb_enabled,
@@ -668,6 +690,8 @@ class FileIOStorageObject(StorageObject):
             doc="Get the current FileIOStorage size in bytes")
     is_block = property(_is_block,
             doc="True if FileIoStorage is backed by a block device instead of a file")
+    aio = property(_aio,
+            doc="True if asynchronous I/O is enabled")
 
     def dump(self):
         d = super(FileIOStorageObject, self).dump()
@@ -675,6 +699,7 @@ class FileIOStorageObject(StorageObject):
         d['wwn'] = self.wwn
         d['dev'] = self.udev_path
         d['size'] = self.size
+        d['aio'] = self.aio
         return d
 
 
@@ -686,7 +711,7 @@ class BlockStorageObject(StorageObject):
     # BlockStorageObject private stuff
 
     def __init__(self, name, dev=None, wwn=None, readonly=False,
-                 write_back=False):
+                 write_back=False, index=None):
         '''
         A BlockIOStorageObject can be instantiated in two ways:
             - B{Creation mode}: If I{dev} is specified, the underlying configFS
@@ -712,14 +737,14 @@ class BlockStorageObject(StorageObject):
         '''
 
         if dev is not None:
-            super(BlockStorageObject, self).__init__(name, 'create')
+            super(BlockStorageObject, self).__init__(name, 'create', index)
             try:
                 self._configure(dev, wwn, readonly)
             except:
                 self.delete()
                 raise
         else:
-            super(BlockStorageObject, self).__init__(name, 'lookup')
+            super(BlockStorageObject, self).__init__(name, 'lookup', index)
 
     def _configure(self, dev, wwn, readonly):
         self._check_self()
@@ -787,7 +812,7 @@ class UserBackedStorageObject(StorageObject):
     '''
 
     def __init__(self, name, config=None, size=None, wwn=None,
-                 hw_max_sectors=None):
+                 hw_max_sectors=None, control=None, index=None):
         '''
         @param name: The name of the UserBackedStorageObject.
         @type name: string
@@ -800,6 +825,9 @@ class UserBackedStorageObject(StorageObject):
         @type wwn: string
         @hw_max_sectors: Max sectors per command limit to export to initiators.
         @type hw_max_sectors: int
+        @control: String of control=value tuples separate by a ',' that will
+            passed to the kernel control file.
+        @type: string
         @return: A UserBackedStorageObject object.
         '''
 
@@ -810,16 +838,16 @@ class UserBackedStorageObject(StorageObject):
             if '/' not in config:
                 raise RTSLibError("'config' must contain a '/' separating subtype "
                                   "from its configuration string")
-            super(UserBackedStorageObject, self).__init__(name, 'create')
+            super(UserBackedStorageObject, self).__init__(name, 'create', index)
             try:
-                self._configure(config, size, wwn, hw_max_sectors)
+                self._configure(config, size, wwn, hw_max_sectors, control)
             except:
                 self.delete()
                 raise
         else:
-            super(UserBackedStorageObject, self).__init__(name, 'lookup')
+            super(UserBackedStorageObject, self).__init__(name, 'lookup', index)
 
-    def _configure(self, config, size, wwn, hw_max_sectors):
+    def _configure(self, config, size, wwn, hw_max_sectors, control):
         self._check_self()
 
         if ':' in config:
@@ -828,6 +856,8 @@ class UserBackedStorageObject(StorageObject):
         self._control("dev_size=%d" % size)
         if hw_max_sectors is not None:
             self._control("hw_max_sectors=%s" % hw_max_sectors)
+        if control is not None:
+            self._control(control)
         self._enable()
 
         super(UserBackedStorageObject, self)._configure(wwn)
@@ -840,6 +870,24 @@ class UserBackedStorageObject(StorageObject):
         self._check_self()
         return int(self._parse_info('HwMaxSectors'))
 
+    def _get_control_tuples(self):
+        self._check_self()
+        tuples = []
+        # 1. max_data_area_mb
+        val = self._parse_info('MaxDataAreaMB')
+        if val != "NULL":
+            tuples.append("max_data_area_mb=%s" % val)
+        val = self.get_attribute('hw_block_size')
+        if val != "NULL":
+            tuples.append("hw_block_size=%s" % val)
+        # 3. data_pages_per_blk
+        val = self._parse_info('DataPagesPerBlk')
+        if val != "NULL":
+            tuples.append("data_pages_per_blk=%s" % val)
+        # 4. add next ...
+
+        return ",".join(tuples)
+
     def _get_config(self):
         self._check_self()
         val = self._parse_info('Config')
@@ -853,6 +901,8 @@ class UserBackedStorageObject(StorageObject):
 
     hw_max_sectors = property(_get_hw_max_sectors,
             doc="Get the max sectors per command.")
+    control_tuples = property(_get_control_tuples,
+            doc="Get the comma separated string containing control=value tuples.")
     size = property(_get_size,
             doc="Get the size in bytes.")
     config = property(_get_config,
@@ -866,6 +916,7 @@ class UserBackedStorageObject(StorageObject):
         d['size'] = self.size
         d['config'] = self.config
         d['hw_max_sectors'] = self.hw_max_sectors
+        d['control'] = self.control_tuples
 
         return d
 
@@ -918,15 +969,16 @@ class _Backstore(CFSNode):
     Created by storageobject ctor before SO configfs entry.
     """
 
-    def __init__(self, name, storage_object_cls, mode):
+    def __init__(self, name, storage_object_cls, mode, index=None):
         super(_Backstore, self).__init__()
         self._so_cls = storage_object_cls
         self._plugin = bs_params[self._so_cls]['name']
 
         dirp = bs_params[self._so_cls].get("alt_dirprefix", self._plugin)
 
+        # if the caller knows the index then skip the cache
         global bs_cache
-        if not bs_cache:
+        if index is None and not bs_cache:
             for dir in glob.iglob("%s/core/*_*/*/" % self.configfs_dir):
                 parts = dir.split("/")
                 bs_name = parts[-2]
@@ -934,19 +986,23 @@ class _Backstore(CFSNode):
                 current_key = "%s/%s" % (bs_dirp, bs_name)
                 bs_cache[current_key] = int(bs_index)
 
-        # mapping in cache?
         self._lookup_key = "%s/%s" % (dirp, name)
-        self._index = bs_cache.get(self._lookup_key, None)
+        if index is None:
+            self._index = bs_cache.get(self._lookup_key, None)
+            if self._index != None and mode == 'create':
+                raise RTSLibError("Storage object %s/%s exists" %
+                                  (self._plugin, name))
+        else:
+            self._index = int(index)
 
-        if self._index != None and mode == 'create':
-            raise RTSLibError("Storage object %s/%s exists" %
-                              (self._plugin, name))
-        elif self._index == None:
+        if self._index == None:
             if mode == 'lookup':
                 raise RTSLibNotInCFS("Storage object %s/%s not found" %
                                      (self._plugin, name))
             else:
                 # Allocate new index value
+                lkfd = open(lock_file, 'w+')
+                fcntl.flock(lkfd, fcntl.LOCK_EX)
                 indexes = set(bs_cache.values())
                 for index in range(1048576):
                     if index not in indexes:
@@ -954,7 +1010,9 @@ class _Backstore(CFSNode):
                         bs_cache[self._lookup_key] = self._index
                         break
                 else:
+                    fcntl.flock(lkfd, fcntl.LOCK_UN)
                     raise RTSLibError("No available backstore index")
+                fcntl.flock(lkfd, fcntl.LOCK_UN)
 
         self._path = "%s/core/%s_%d" % (self.configfs_dir,
                                         dirp,
@@ -962,12 +1020,14 @@ class _Backstore(CFSNode):
         try:
             self._create_in_cfs_ine(mode)
         except:
-            del bs_cache[self._lookup_key]
+            if self._lookup_key in bs_cache:
+                del bs_cache[self._lookup_key]
             raise
 
     def delete(self):
         super(_Backstore, self).delete()
-        del bs_cache[self._lookup_key]
+        if self._lookup_key in bs_cache:
+            del bs_cache[self._lookup_key]
 
     def _get_index(self):
         return self._index
@@ -975,8 +1035,11 @@ class _Backstore(CFSNode):
     def _parse_info(self, key):
         self._check_self()
         info = fread("%s/hba_info" % self.path)
-        return re.search(".*%s: ([^: ]+).*" \
-                         % key, ' '.join(info.split())).group(1)
+        try:
+            return re.search(".*%s: ([^: ]+).*" \
+                             % key, ' '.join(info.split())).group(1)
+        except AttributeError:
+            return None
 
     def _get_version(self):
         self._check_self()
